{% import 'kiara/render/models/macros.html' as model_macros %}

<table>
    <tr>
        <td>Documentation</td>
        <td>{{ instance.documentation.full_doc | markdown | safe }}</td>
    </tr>
    <tr>
        <td>Inputs</td>
        <td>{{ model_macros.field_infos(field_infos=instance.input_fields) }}</td>
    </tr>
    <tr>
        <td>Outputs</td>
        <td>{{ model_macros.field_infos(field_infos=instance.output_fields) }}</td>
    </tr>
</table>
